use std::collections::{HashMap, VecDeque};
use std::io::{Read, Write};
use std::iter;
use std::os::unix::net::UnixStream;
use std::process::{Command, Stdio};
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::Arc;
use std::thread::{self, JoinHandle};
use std::time::Duration;

use color_eyre::eyre::{self, Context};
use color_eyre::Result;
use crossbeam::deque::{Injector, Stealer, Worker};
use tree_sitter_highlight::{HighlightConfiguration, HighlightEvent, Highlighter};

use crate::kakoune::editor_quote;
use crate::Request;

type Task = Result<UnixStream>;

pub struct TaskScheduler {
    injector: Arc<Injector<Task>>,
    terminate: Arc<AtomicBool>,
    threads: Vec<JoinHandle<()>>,
}

pub struct TaskContext {
    worker: Worker<Task>,
    injector: Arc<Injector<Task>>,
    stealers: Arc<Vec<Stealer<Task>>>,
    terminate: Arc<AtomicBool>,
    highlighter: Highlighter,
    highlight_config: Arc<HighlightConfiguration>,
    tokens: Arc<HashMap<String, String>>,
}

pub fn make_highlighter_config<S: AsRef<str>>(
    names: impl AsRef<[S]>,
) -> Result<HighlightConfiguration> {
    let mut config = HighlightConfiguration::new(
        tree_sitter_rust::language(),
        tree_sitter_rust::HIGHLIGHT_QUERY,
        tree_sitter_rust::INJECTIONS_QUERY,
        tree_sitter_rust::LOCALS_QUERY,
    )
    .wrap_err("Invalid highlighter config")?;

    config.configure(names.as_ref());

    Ok(config)
}

impl TaskScheduler {
    pub fn new(workers: u8, tokens: HashMap<String, String>) -> Result<Self> {
        let terminate = Arc::new(AtomicBool::new(false));

        let injector = Arc::new(Injector::new());
        let workers: Vec<_> = (0..workers).map(|_| Worker::new_fifo()).collect();
        let stealers: Vec<_> = workers.iter().map(|w| w.stealer()).collect();
        let stealers = Arc::new(stealers);

        let names: Vec<_> = tokens.keys().collect();
        tracing::debug!("Highlighter tokens: {:?}", names);

        let highlight_config = make_highlighter_config(names)?;
        let highlight_config = Arc::new(highlight_config);
        let tokens = Arc::new(tokens);

        let threads = workers
            .into_iter()
            .enumerate()
            .map(|(i, worker)| {
                let ctx = TaskContext {
                    worker,
                    injector: injector.clone(),
                    stealers: stealers.clone(),
                    terminate: terminate.clone(),
                    highlighter: Highlighter::new(),
                    highlight_config: highlight_config.clone(),
                    tokens: tokens.clone(),
                };

                thread::Builder::new()
                    .name(format!("worker-{}", i))
                    .spawn(|| thread_handler(ctx))
                    .map_err(From::from)
            })
            .collect::<Result<Vec<_>>>()
            .wrap_err("Failed to spawn worker threads")?;

        tracing::info!("Started {} worker threads", threads.len());

        Ok(Self {
            injector,
            terminate,
            threads,
        })
    }

    pub fn schedule(&mut self, task: Task) {
        self.injector.push(task)
    }

    pub fn terminate(self) {
        self.terminate.store(true, Ordering::Relaxed);

        for handle in self.threads {
            if let Err(err) = handle.join() {
                tracing::error!("Worker thread panicked: {:?}", err);
            }
        }
    }
}

#[tracing::instrument]
fn find_task<T>(local: &Worker<T>, global: &Injector<T>, stealers: &[Stealer<T>]) -> Option<T> {
    local.pop().or_else(|| {
        iter::repeat_with(|| {
            global
                .steal_batch_and_pop(local)
                .or_else(|| stealers.iter().map(|s| s.steal()).collect())
        })
        .find(|s| !s.is_retry())
        .and_then(|s| s.success())
    })
}

#[tracing::instrument(skip_all)]
fn thread_handler(mut ctx: TaskContext) {
    loop {
        let task = 'find_task: loop {
            if ctx.terminate.load(Ordering::Relaxed) {
                return;
            }

            if let Some(task) = find_task(&ctx.worker, &ctx.injector, &ctx.stealers) {
                break 'find_task task;
            }

            thread::sleep(Duration::from_millis(50));
        };

        if let Err(err) = handle_connection(&mut ctx, task) {
            tracing::error!("{:?}", err);
        }
    }
}

#[tracing::instrument(skip_all)]
fn handle_connection(ctx: &mut TaskContext, task: Task) -> Result<()> {
    let mut stream = task.wrap_err("Failed to receive client connection")?;

    let mut buf = String::new();
    stream
        .read_to_string(&mut buf)
        .wrap_err("Failed to read from connection")?;
    let req: Request = toml::from_str(&buf).wrap_err("Failed to parse request")?;

    tracing::info!("Received request");
    tracing::debug!(?req);

    let response = highlight_content(
        &mut ctx.highlighter,
        &ctx.highlight_config,
        &ctx.tokens,
        req.content.as_bytes(),
        req.timestamp,
    )
    .unwrap_or_else(|err| format!("fail {}", editor_quote(format!("{}", err))));

    let mut child = Command::new("kak")
        .args(["-p", &req.session])
        .stdin(Stdio::piped())
        .stdout(Stdio::null())
        .stderr(Stdio::null())
        .spawn()
        .wrap_err("Failed to spawn Kakoune command")?;

    if let Some(stdin) = child.stdin.as_mut() {
        let command = format!(
            "evaluate-commands -client {} -verbatim -- {}",
            req.client, response
        );

        tracing::info!("Writing response");
        tracing::debug!(command);

        stdin
            .write_all(command.as_bytes())
            .wrap_err("Failed to write to Kakoune stdin")?;
    } else {
        eyre::bail!("Failed to get stdin for Kakoune command");
    }

    Ok(())
}

// Tree-sitter operates on byte offsets, while Kakoune's range specs use
// (row, column) tuples. Calculating those on the fly is very expensive, as it requires
// iterating the whole content to find line breaks.
// So instead, generate a map once where any byte offset can be looked up later.
fn build_range_map(content: &[u8]) -> Vec<(usize, usize)> {
    let mut map = Vec::with_capacity(content.len());

    // Kakoune's line indices are 1-based
    let mut row = 1;
    let mut column = 1;

    for byte in content.iter() {
        map.push((row, column));

        if *byte == b'\n' {
            row += 1;
            column = 1;
        } else {
            column += 1;
        }
    }

    map
}

#[tracing::instrument(skip(highlighter, highlight_config, tokens), fields(
    content_len = content.len(),
))]
pub fn highlight_content(
    highlighter: &mut Highlighter,
    highlight_config: &HighlightConfiguration,
    tokens: &HashMap<String, String>,
    content: &[u8],
    timestamp: u64,
) -> Result<String> {
    // By building vectors for both keys and values with the same order, we can
    // use the highlighter's returned index to retrieve the face directly, rather
    // than having to perform a second lookup in the hash map.
    let mut names = Vec::with_capacity(tokens.len());
    let mut faces = Vec::with_capacity(tokens.len());
    for (name, face) in tokens.iter() {
        names.push(name);
        faces.push(face);
    }

    let highlights = highlighter
        .highlight(highlight_config, content, None, |lang| {
            if lang == "rust" {
                Some(highlight_config)
            } else {
                None
            }
        })
        .wrap_err("Failed to highlight content")?;

    let mut stack = VecDeque::with_capacity(64);
    let range_map = build_range_map(content);
    // We're going to do lots of small pushes to this String, which would trigger
    // lots of incremental re-allocations. To avoid that, start with a capacity that should
    // hoefully be somewhat close to, but bigger than the final size.
    let mut response = String::with_capacity(content.len() * 2);
    response.push_str("set-option buffer kak_highlight_ranges ");
    response.push_str(&timestamp.to_string());

    for res in highlights {
        match res? {
            HighlightEvent::Source { start, end } => {
                if let Some(index) = stack.back() {
                    // Tree-sitter actually returns the byte position after the token
                    // as `end` here.
                    let end = end.saturating_sub(1);

                    let (start_row, start_column) = range_map[start];
                    let (end_row, end_column) = range_map[end];
                    let face: &String = faces[*index];

                    response.push(' ');

                    tracing::trace!(index, start_offset = start, end_offset = end, face);

                    response.push_str(&start_row.to_string());
                    response.push('.');
                    response.push_str(&start_column.to_string());
                    response.push(',');
                    response.push_str(&end_row.to_string());
                    response.push('.');
                    response.push_str(&end_column.to_string());
                    response.push('|');
                    response.push_str(face);
                }
            }
            HighlightEvent::HighlightStart(index) => {
                stack.push_back(index.0);
            }
            HighlightEvent::HighlightEnd => {
                // Tree-sitter shouldn't call this when there is nothing on the stack,
                // but it wouldn't matter anyways.
                let _ = stack.pop_back();
            }
        }
    }

    Ok(response)
}

#[cfg(test)]
mod test {
    use std::collections::HashMap;

    use tree_sitter_highlight::Highlighter;

    use crate::daemon::worker::make_highlighter_config;

    use super::highlight_content;

    fn make_tokens() -> HashMap<String, String> {
        [
            "constructor",
            "function",
            "function.macro",
            "function.method",
            "keyword",
            "module",
            "operator",
            "punctuation",
            "string",
            "type",
            "variable",
            "variable.parameter",
        ]
        .into_iter()
        .fold(HashMap::new(), |mut map, val| {
            map.insert(String::from(val), String::from(val));
            map
        })
    }

    fn run_test(content: impl AsRef<str>, specs: &[&str]) {
        let timestamp = 0;
        let tokens = make_tokens();
        let names: Vec<_> = tokens.keys().collect();

        let mut highlighter = Highlighter::new();
        let highlight_config = make_highlighter_config(&names).unwrap();

        let response = highlight_content(
            &mut highlighter,
            &highlight_config,
            &tokens,
            content.as_ref().as_bytes(),
            timestamp,
        )
        .unwrap();

        assert_eq!(
            response,
            format!(
                "set-option buffer kak_highlight_ranges {timestamp} {}",
                specs.join(" ")
            )
            .trim()
        );
    }

    #[test]
    fn empty_content() {
        let content = String::new();
        let specs = [];
        run_test(content, &specs);
    }

    #[test]
    fn main_rs() {
        let content = r#"fn main() {
    println!("Hello, world!");
}
"#;

        let specs = vec![
            "1.1,1.2|keyword",
            "1.4,1.7|function",
            "1.8,1.8|punctuation",
            "1.9,1.9|punctuation",
            "1.11,1.11|punctuation",
            "2.5,2.11|function.macro",
            "2.12,2.12|function.macro",
            "2.13,2.13|punctuation",
            "2.14,2.28|string",
            "2.29,2.29|punctuation",
            "2.30,2.30|punctuation",
            "3.1,3.1|punctuation",
        ];

        run_test(content, &specs);
    }
}
